@@ -1,14 +1,16 @@
 """
 Encoders for encoding categorical variables and scaling continuous data.
 """
-from typing import Dict, Iterable, List, Tuple, Union
+from typing import Callable, Dict, Iterable, List, Tuple, Union
 import warnings
 
 import numpy as np
 import pandas as pd
+from pandas.core.algorithms import isin
 from sklearn.base import BaseEstimator, TransformerMixin
 import torch
 import torch.nn.functional as F
+from torch.nn.utils import rnn
 
 
 class NaNLabelEncoder(BaseEstimator, TransformerMixin):
@@ -80,12 +82,19 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
         self.classes_vector_ = np.array(list(self.classes_.keys()))
         return self
 
-    def transform(self, y: Iterable) -> Union[torch.Tensor, np.ndarray]:
+    def transform(
+        self, y: Iterable, return_norm: bool = False, target_scale=None, ignore_na: bool = False
+    ) -> Union[torch.Tensor, np.ndarray]:
         """
         Encode iterable with integers.
 
         Args:
             y (Iterable): iterable to encode
+            return_norm: only exists for compatability with other encoders - returns a tuple if true.
+            target_scale: only exists for compatability with other encoders - has no effect.
+            ignore_na (bool): if to ignore na values and map them to zeros
+                (this is different to `add_nan=True` option which maps ONLY NAs to zeros
+                while this options maps the first class and NAs to zeros)
 
         Returns:
             Union[torch.Tensor, np.ndarray]: returns encoded data as torch tensor or numpy array depending on input type
@@ -102,13 +111,26 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
             encoded = [self.classes_.get(v, 0) for v in y]
 
         else:
-            encoded = [self.classes_[v] for v in y]
+            if ignore_na:
+                na_fill_value = next(iter(self.classes_.values()))
+                encoded = [self.classes_.get(v, na_fill_value) for v in y]
+            else:
+                try:
+                    encoded = [self.classes_[v] for v in y]
+                except KeyError as e:
+                    raise KeyError(
+                        f"Unknown category '{e.args[0]}' encountered. Set `add_nan=True` to allow unknown categories"
+                    )
 
         if isinstance(y, torch.Tensor):
             encoded = torch.tensor(encoded, dtype=torch.long, device=y.device)
         else:
             encoded = np.array(encoded)
-        return encoded
+
+        if return_norm:
+            return encoded, self.get_parameters()
+        else:
+            return encoded
 
     def inverse_transform(self, y: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
         """
@@ -144,56 +166,79 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
         """
         return data["prediction"]
 
+    def get_parameters(self, groups=None, group_names=None) -> np.ndarray:
+        """
+        Get fitted scaling parameters for a given group.
+
+        All parameters are unused - exists for compatability.
+
+        Returns:
+            np.ndarray: zero array.
+        """
+        return np.zeros(2, dtype=np.float)
+
+
+def _plus_one(x):
+    return x + 1
+
+
+def _identity(x):
+    return x
+
+
+def _clamp_zero(x):
+    return x.clamp(0.0)
+
 
 class TorchNormalizer(BaseEstimator, TransformerMixin):
     """
     Basic target transformer that can be fit also on torch tensors.
     """
 
+    # transformation and inverse transformation
+    TRANSFORMATIONS = {
+        "log": (torch.log, torch.exp),
+        "log1p": (torch.log1p, torch.exp),
+        "logit": (torch.logit, torch.sigmoid),
+        "softplus": (_plus_one, F.softplus),
+        "relu": (_identity, _clamp_zero),
+    }
+
     def __init__(
         self,
         method: str = "standard",
         center: bool = True,
-        log_scale: Union[bool, float] = False,
-        log_zero_value: float = 0.0,
-        coerce_positive: Union[float, bool] = None,
+        transformation: Union[str, Tuple[Callable, Callable]] = None,
         eps: float = 1e-8,
     ):
         """
         Initialize
 
         Args:
-            method (str, optional): method to rescale series. Either "standard" (standard scaling) or "robust"
-                (scale using quantiles 0.25-0.75). Defaults to "standard".
+            method (str, optional): method to rescale series. Either "identity", "standard" (standard scaling)
+                or "robust" (scale using quantiles 0.25-0.75). Defaults to "standard".
             center (bool, optional): If to center the output to zero. Defaults to True.
-            log_scale (bool, optional): If to take log of values. Defaults to False. Defaults to False.
-            log_zero_value (float, optional): Value to map 0 to for ``log_scale=True`` or in softplus. Defaults to 0.0
-            coerce_positive (Union[bool, float, str], optional): If to coerce output to positive. Valid values:
-                * None, i.e. is automatically determined and might change to True if all values are >= 0 (Default).
-                * True, i.e. output is clamped at 0.
-                * False, i.e. values are not coerced
-                * float, i.e. softmax is applied with beta = coerce_positive.
-            eps (float, optional): Number for numerical stability of calcualtions. Defaults to 1e-8.
+            transformation (Union[str, Tuple[Callable, Callable]] optional): Transform values before
+                applying normalizer. Available options are
+
+                * None (default): No transformation of values
+                * log: Estimate in log-space leading to a multiplicative model
+                * logp1: Estimate in log-space but add 1 to values before transforming for stability
+                    (e.g. if many small values <<1 are present).
+                    Note, that inverse transform is still only `torch.exp()` and not `torch.expm1()`.
+                * logit: Apply logit transformation on values that are between 0 and 1
+                * softplus: Apply softplus to output (inverse transformation) and x + 1 to input (transformation)
+                * relu: Apply max(0, x) to output
+                * Tuple[Callable, Callable] of PyTorch functions that transforms and inversely transforms values.
+
+            eps (float, optional): Number for numerical stability of calculations.
+                Defaults to 1e-8.
         """
         self.method = method
-        assert method in ["standard", "robust"], f"method has invalid value {method}"
+        assert method in ["standard", "robust", "identity"], f"method has invalid value {method}"
         self.center = center
         self.eps = eps
-
-        # set log scale
-        self.log_zero_value = np.exp(log_zero_value)
-        self.log_scale = log_scale
-
-        # check if coerce positive should be determined automatically
-        if coerce_positive is None:
-            if log_scale:
-                coerce_positive = False
-        else:
-            assert not (self.log_scale and coerce_positive), (
-                "log scale means that output is transformed to a positive number by default while coercing positive"
-                " will apply softmax function - decide for either one or the other"
-            )
-        self.coerce_positive = coerce_positive
+        self.transformation = transformation
 
     def get_parameters(self, *args, **kwargs) -> torch.Tensor:
         """
@@ -202,25 +247,58 @@ class TorchNormalizer(BaseEstimator, TransformerMixin):
         Returns:
             torch.Tensor: First element is center of data and second is scale
         """
-        return torch.tensor([self.center_, self.scale_])
+        return torch.stack([torch.as_tensor(self.center_), torch.as_tensor(self.scale_)], dim=-1)
 
-    def _preprocess_y(self, y: Union[pd.Series, np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
+    def preprocess(
+        self, y: Union[pd.Series, pd.DataFrame, np.ndarray, torch.Tensor]
+    ) -> Union[np.ndarray, torch.Tensor]:
         """
         Preprocess input data (e.g. take log).
 
-        Can set coerce positive to a value if it was set to None and log_scale to False.
+        Uses ``transform`` attribute to determine how to apply transform.
 
         Returns:
             Union[np.ndarray, torch.Tensor]: return rescaled series with type depending on input type
         """
-        if self.coerce_positive is None and not self.log_scale:
-            self.coerce_positive = (y >= 0).all()
+        if self.transformation is None:
+            return y
 
-        if self.log_scale:
-            if isinstance(y, torch.Tensor):
-                y = torch.log(y + self.log_zero_value)
-            else:
-                y = np.log(y + self.log_zero_value)
+        # protect against numerical instabilities
+        if isinstance(self.transformation, str) and self.transformation == "logit":
+            # need to apply eps slightly differently
+            y = y / (1 + 2 * self.eps) + self.eps
+        else:
+            y = y + self.eps
+
+        if isinstance(y, torch.Tensor):
+            y = self.TRANSFORMATIONS.get(self.transformation, self.transformation)[0](y)
+        else:
+            # convert first to tensor, then transform and then convert to numpy array
+            if isinstance(y, (pd.Series, pd.DataFrame)):
+                y = y.to_numpy()
+            y = torch.as_tensor(y)
+            y = self.TRANSFORMATIONS.get(self.transformation, self.transformation)[0](y)
+            y = np.asarray(y)
+        return y
+
+    def inverse_preprocess(self, y: Union[pd.Series, np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
+        """
+        Inverse preprocess re-scaled data (e.g. take exp).
+
+        Uses ``transform`` attribute to determine how to apply inverse transform.
+
+        Returns:
+            Union[np.ndarray, torch.Tensor]: return rescaled series with type depending on input type
+        """
+        if self.transformation is None:
+            pass
+        elif isinstance(y, torch.Tensor):
+            y = self.TRANSFORMATIONS.get(self.transformation, self.transformation)[1](y)
+        else:
+            # convert first to tensor, then transform and then convert to numpy array
+            y = torch.as_tensor(y)
+            y = self.TRANSFORMATIONS.get(self.transformation, self.transformation)[1](y)
+            y = np.asarray(y)
         return y
 
     def fit(self, y: Union[pd.Series, np.ndarray, torch.Tensor]):
@@ -233,30 +311,57 @@ class TorchNormalizer(BaseEstimator, TransformerMixin):
         Returns:
             TorchNormalizer: self
         """
-        y = self._preprocess_y(y)
+        y = self.preprocess(y)
 
-        if self.method == "standard":
+        if self.method == "identity":
             if isinstance(y, torch.Tensor):
-                self.center_ = torch.mean(y)
-                self.scale_ = torch.std(y) / (self.center_ + self.eps)
+                self.center_ = torch.zeros(y.size()[:-1])
+                self.scale_ = torch.ones(y.size()[:-1])
+            elif isinstance(y, (np.ndarray, pd.Series, pd.DataFrame)):
+                self.center_ = np.zeros(y.shape[:-1])
+                self.scale_ = np.ones(y.shape[:-1])
+            else:
+                self.center_ = 0.0
+                self.scale_ = 1.0
+
+        elif self.method == "standard":
+            if isinstance(y, torch.Tensor):
+                self.center_ = torch.mean(y, dim=-1)
+                self.scale_ = torch.std(y, dim=-1) + self.eps
+            elif isinstance(y, np.ndarray):
+                self.center_ = np.mean(y, axis=-1)
+                self.scale_ = np.std(y, axis=-1) + self.eps
             else:
                 self.center_ = np.mean(y)
-                self.scale_ = np.std(y) / (self.center_ + self.eps)
+                self.scale_ = np.std(y) + self.eps
 
         elif self.method == "robust":
             if isinstance(y, torch.Tensor):
-                self.center_ = torch.median(y)
-                q_75 = y.kthvalue(int(len(y) * 0.75)).values
-                q_25 = y.kthvalue(int(len(y) * 0.25)).values
+                self.center_ = torch.median(y, dim=-1).values
+                q_75 = y.kthvalue(int(len(y) * 0.75), dim=-1).values
+                q_25 = y.kthvalue(int(len(y) * 0.25), dim=-1).values
+            elif isinstance(y, np.ndarray):
+                self.center_ = np.median(y, axis=-1)
+                q_75 = np.percentiley(y, 75, axis=-1)
+                q_25 = np.percentiley(y, 25, axis=-1)
             else:
                 self.center_ = np.median(y)
                 q_75 = np.percentiley(y, 75)
                 q_25 = np.percentiley(y, 25)
-            self.scale_ = (q_75 - q_25) / (self.center_ + self.eps) / 2.0
+            self.scale_ = (q_75 - q_25) / 2.0 + self.eps
+        if not self.center:
+            self.scale_ = self.center_
+            if isinstance(y, torch.Tensor):
+                self.center_ = torch.zeros_like(self.center_)
+            else:
+                self.center_ = np.zeros_like(self.center_)
         return self
 
     def transform(
-        self, y: Union[pd.Series, np.ndarray, torch.Tensor], return_norm: bool = False
+        self,
+        y: Union[pd.Series, np.ndarray, torch.Tensor],
+        return_norm: bool = False,
+        target_scale: torch.Tensor = None,
     ) -> Union[Tuple[Union[np.ndarray, torch.Tensor], np.ndarray], Union[np.ndarray, torch.Tensor]]:
         """
         Rescale data.
@@ -264,22 +369,28 @@ class TorchNormalizer(BaseEstimator, TransformerMixin):
         Args:
             y (Union[pd.Series, np.ndarray, torch.Tensor]): input data
             return_norm (bool, optional): [description]. Defaults to False.
+            target_scale (torch.Tensor): target scale to use instead of fitted center and scale
 
         Returns:
             Union[Tuple[Union[np.ndarray, torch.Tensor], np.ndarray], Union[np.ndarray, torch.Tensor]]: rescaled
                 data with type depending on input type. returns second element if ``return_norm=True``
         """
-        if self.log_scale:
-            if isinstance(y, torch.Tensor):
-                y = (y + self.log_zero_value + self.eps).log()
-            else:
-                y = np.log(y + self.log_zero_value + self.eps)
-        if self.center:
-            y = (y / (self.center_ + self.eps) - 1) / (self.scale_ + self.eps)
-        else:
-            y = y / (self.center_ + self.eps)
+        y = self.preprocess(y)
+        # get center and scale
+        if target_scale is None:
+            target_scale = self.get_parameters().numpy()[None, :]
+        center = target_scale[..., 0]
+        scale = target_scale[..., 1]
+        if y.ndim > center.ndim:  # multiple batches -> expand size
+            center = center.view(*center.size(), *(1,) * (y.ndim - center.ndim))
+            scale = scale.view(*scale.size(), *(1,) * (y.ndim - scale.ndim))
+
+        # transform
+        y = (y - center) / scale
+
+        # return with center and scale or without
         if return_norm:
-            return y, self.get_parameters().numpy()[None, :]
+            return y, target_scale
         else:
             return y
 
@@ -315,28 +426,22 @@ class TorchNormalizer(BaseEstimator, TransformerMixin):
             norm = norm.unsqueeze(-1)
 
         # transform
-        if self.center:
-            y_normed = (data["prediction"] * norm[:, 1, None] + 1) * norm[:, 0, None]
-        else:
-            y_normed = data["prediction"] * norm[:, 0, None]
-        if self.log_scale:
-            y_normed = (y_normed.exp() - self.log_zero_value).clamp_min(0.0)
-        elif isinstance(self.coerce_positive, bool) and self.coerce_positive:
-            y_normed = y_normed.clamp_min(0.0)
-        elif isinstance(self.coerce_positive, float):
-            y_normed = F.softplus(y_normed, beta=float(self.coerce_positive))
+        y = data["prediction"] * norm[:, 1, None] + norm[:, 0, None]
+
+        y = self.inverse_preprocess(y)
 
         # return correct shape
-        if data["prediction"].ndim == 1 and y_normed.ndim > 1:
-            y_normed = y_normed.squeeze(0)
-        return y_normed
+        if data["prediction"].ndim == 1 and y.ndim > 1:
+            y = y.squeeze(0)
+        return y
 
 
 class EncoderNormalizer(TorchNormalizer):
     """
     Special Normalizer that is fit on each encoding sequence.
 
-    If passed as target normalizer, this transformer will be fitted on each encoder sequence separately.
+    If used, this transformer will be fitted on each encoder sequence separately.
+    This normalizer can be particularly useful as target normalizer.
     """
 
     pass
@@ -350,16 +455,13 @@ class GroupNormalizer(TorchNormalizer):
     also to normalize any other variable.
     """
 
-    # todo: allow window (exp weighted), different methods such as quantile for robust scaling
     def __init__(
         self,
         method: str = "standard",
         groups: List[str] = [],
         center: bool = True,
         scale_by_group: bool = False,
-        log_scale: Union[bool, float] = False,
-        log_zero_value: float = 0.0,
-        coerce_positive: Union[float, bool] = None,
+        transformation: Union[str, Tuple[Callable, Callable]] = None,
         eps: float = 1e-8,
     ):
         """
@@ -372,23 +474,28 @@ class GroupNormalizer(TorchNormalizer):
             center (bool, optional): If to center the output to zero. Defaults to True.
             scale_by_group (bool, optional): If to scale the output by group, i.e. norm is calculated as
                 ``(group1_norm * group2_norm * ...) ^ (1 / n_groups)``. Defaults to False.
-            log_scale (bool, optional): If to take log of values. Defaults to False. Defaults to False.
-            log_zero_value (float, optional): Value to map 0 to for ``log_scale=True`` or in softplus. Defaults to 0.0
-            coerce_positive (Union[bool, float, str], optional): If to coerce output to positive. Valid values:
-                * None, i.e. is automatically determined and might change to True if all values are >= 0 (Default).
-                * True, i.e. output is clamped at 0.
-                * False, i.e. values are not coerced
-                * float, i.e. softmax is applied with beta = coerce_positive.
-            eps (float, optional): Number for numerical stability of calcualtions. Defaults to 1e-8.
+            transformation (Union[str, Tuple[Callable, Callable]] optional): Transform values before
+                applying normalizer. Available options are
+
+                * None (default): No transformation of values
+                * log: Estimate in log-space leading to a multiplicative model
+                * logp1: Estimate in log-space but add 1 to values before transforming for stability
+                    (e.g. if many small values <<1 are present).
+                    Note, that inverse transform is still only `torch.exp()` and not `torch.expm1()`.
+                * logit: Apply logit transformation on values that are between 0 and 1
+                * softplus: Apply softplus to output (inverse transformation) and x + 1 to input (transformation)
+                * relu: Apply max(0, x) to output
+                * Tuple[Callable, Callable] of PyTorch functions that transforms and inversely transforms values.
+
+            eps (float, optional): Number for numerical stability of calcualtions.
+                Defaults to 1e-8. For count data, 1.0 is recommended.
         """
         self.groups = groups
         self.scale_by_group = scale_by_group
         super().__init__(
             method=method,
             center=center,
-            log_scale=log_scale,
-            log_zero_value=log_zero_value,
-            coerce_positive=coerce_positive,
+            transformation=transformation,
             eps=eps,
         )
 
@@ -403,15 +510,20 @@ class GroupNormalizer(TorchNormalizer):
         Returns:
             self
         """
-        y = self._preprocess_y(y)
+        y = self.preprocess(y)
         if len(self.groups) == 0:
             assert not self.scale_by_group, "No groups are defined, i.e. `scale_by_group=[]`"
             if self.method == "standard":
-                mean = np.mean(y)
-                self.norm_ = mean, np.std(y) / (mean + self.eps)
+                self.norm_ = [np.mean(y), np.std(y) + self.eps]  # center and scale
             else:
                 quantiles = np.quantile(y, [0.25, 0.5, 0.75])
-                self.norm_ = quantiles[1], (quantiles[2] - quantiles[0]) / (quantiles[1] + self.eps)
+                self.norm_ = [
+                    quantiles[1],
+                    (quantiles[2] - quantiles[0]) / 2.0 + self.eps,
+                ]  # center and scale
+            if not self.center:
+                self.norm_[1] = self.norm_[0]
+                self.norm_[0] = 0.0
 
         elif self.scale_by_group:
             if self.method == "standard":
@@ -419,8 +531,8 @@ class GroupNormalizer(TorchNormalizer):
                     g: X[[g]]
                     .assign(y=y)
                     .groupby(g, observed=True)
-                    .agg(mean=("y", "mean"), scale=("y", "std"))
-                    .assign(scale=lambda x: x.scale / (x["mean"] + self.eps))
+                    .agg(center=("y", "mean"), scale=("y", "std"))
+                    .assign(center=lambda x: x["center"], scale=lambda x: x.scale + self.eps)
                     for g in self.groups
                 }
             else:
@@ -431,12 +543,20 @@ class GroupNormalizer(TorchNormalizer):
                     .y.quantile([0.25, 0.5, 0.75])
                     .unstack(-1)
                     .assign(
-                        median=lambda x: x[0.5] + self.eps,
-                        scale=lambda x: (x[0.75] - x[0.25] + self.eps) / (x[0.5] + self.eps),
-                    )[["median", "scale"]]
+                        center=lambda x: x[0.5],
+                        scale=lambda x: (x[0.75] - x[0.25]) / 2.0 + self.eps,
+                    )[["center", "scale"]]
                     for g in self.groups
                 }
             # calculate missings
+            if not self.center:  # swap center and scale
+
+                def swap_parameters(norm):
+                    norm["scale"] = norm["center"]
+                    norm["center"] = 0.0
+                    return norm
+
+                self.norm_ = {g: swap_parameters(norm) for g, norm in self.norm_.items()}
             self.missing_ = {group: scales.median().to_dict() for group, scales in self.norm_.items()}
 
         else:
@@ -445,8 +565,8 @@ class GroupNormalizer(TorchNormalizer):
                     X[self.groups]
                     .assign(y=y)
                     .groupby(self.groups, observed=True)
-                    .agg(mean=("y", "mean"), scale=("y", "std"))
-                    .assign(scale=lambda x: x.scale / (x["mean"] + self.eps))
+                    .agg(center=("y", "mean"), scale=("y", "std"))
+                    .assign(center=lambda x: x["center"], scale=lambda x: x.scale + self.eps)
                 )
             else:
                 self.norm_ = (
@@ -456,10 +576,13 @@ class GroupNormalizer(TorchNormalizer):
                     .y.quantile([0.25, 0.5, 0.75])
                     .unstack(-1)
                     .assign(
-                        median=lambda x: x[0.5] + self.eps,
-                        scale=lambda x: (x[0.75] - x[0.25] + self.eps) / (x[0.5] + self.eps) / 2.0,
-                    )[["median", "scale"]]
+                        center=lambda x: x[0.5],
+                        scale=lambda x: (x[0.75] - x[0.25]) / 2.0 + self.eps,
+                    )[["center", "scale"]]
                 )
+            if not self.center:  # swap center and scale
+                self.norm_["scale"] = self.norm_["center"]
+                self.norm_["center"] = 0.0
             self.missing_ = self.norm_.median().to_dict()
         return self
 
@@ -471,10 +594,7 @@ class GroupNormalizer(TorchNormalizer):
         Returns:
             List[str]: list of names
         """
-        if self.method == "standard":
-            return ["mean", "scale"]
-        else:
-            return ["median", "scale"]
+        return ["center", "scale"]
 
     def fit_transform(
         self, y: pd.Series, X: pd.DataFrame, return_norm: bool = False
@@ -495,12 +615,12 @@ class GroupNormalizer(TorchNormalizer):
 
     def inverse_transform(self, y: pd.Series, X: pd.DataFrame):
         """
-        Rescaling data to original scale - not implemented.
+        Rescaling data to original scale - not implemented - call class with target scale instead.
         """
         raise NotImplementedError()
 
     def transform(
-        self, y: pd.Series, X: pd.DataFrame, return_norm: bool = False
+        self, y: pd.Series, X: pd.DataFrame = None, return_norm: bool = False, target_scale: torch.Tensor = None
     ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
         """
         Scale input data.
@@ -509,21 +629,16 @@ class GroupNormalizer(TorchNormalizer):
             y (pd.Series): data to scale
             X (pd.DataFrame): dataframe with ``groups`` columns
             return_norm (bool, optional): If to return . Defaults to False.
+            target_scale (torch.Tensor): target scale to use instead of fitted center and scale
 
         Returns:
             Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]: Scaled data, if ``return_norm=True``, returns also scales
                 as second element
         """
-        norm = self.get_norm(X)
-        y = self._preprocess_y(y)
-        if self.center:
-            y_normed = (y / (norm[:, 0] + self.eps) - 1) / (norm[:, 1] + self.eps)
-        else:
-            y_normed = y / (norm[:, 0] + self.eps)
-        if return_norm:
-            return y_normed, norm
-        else:
-            return y_normed
+        if target_scale is None:
+            assert X is not None, "either target_scale or X has to be passed"
+            target_scale = self.get_norm(X)
+        return super().transform(y=y, return_norm=return_norm, target_scale=target_scale)
 
     def get_parameters(self, groups: Union[torch.Tensor, list, tuple], group_names: List[str] = None) -> np.ndarray:
         """
@@ -596,3 +711,182 @@ class GroupNormalizer(TorchNormalizer):
         else:
             norm = X[self.groups].set_index(self.groups).join(self.norm_).fillna(self.missing_).to_numpy()
         return norm
+
+
+class MultiNormalizer(TorchNormalizer):
+    """
+    Normalizer for multiple targets.
+
+    This normalizers wraps multiple other normalizers.
+    """
+
+    def __init__(self, normalizers: List[TorchNormalizer]):
+        """
+        Args:
+            normalizers (List[TorchNormalizer]): list of normalizers to apply to targets
+        """
+        self.normalizers = normalizers
+
+    def fit(self, y: Union[pd.DataFrame, np.ndarray, torch.Tensor], X: pd.DataFrame = None):
+        """
+        Fit transformer, i.e. determine center and scale of data
+
+        Args:
+            y (Union[pd.Series, np.ndarray, torch.Tensor]): input data
+
+        Returns:
+            MultiNormalizer: self
+        """
+        if isinstance(y, pd.DataFrame):
+            y = y.to_numpy()
+
+        for idx, normalizer in enumerate(self.normalizers):
+            if isinstance(normalizer, GroupNormalizer):
+                normalizer.fit(y[:, idx], X=X)
+            else:
+                normalizer.fit(y[:, idx])
+
+        return self
+
+    def __getitem__(self, idx: int):
+        """
+        Return normalizer.
+
+        Args:
+            idx (int): metric index
+        """
+        return self.normalizers[idx]
+
+    def __iter__(self):
+        """
+        Iter over normalizers.
+        """
+        return iter(self.normalizers)
+
+    def __len__(self) -> int:
+        """
+        Number of normalizers.
+        """
+        return len(self.normalizers)
+
+    def transform(
+        self,
+        y: Union[pd.DataFrame, np.ndarray, torch.Tensor],
+        X: pd.DataFrame = None,
+        return_norm: bool = False,
+        target_scale: List[torch.Tensor] = None,
+    ) -> Union[List[Tuple[Union[np.ndarray, torch.Tensor], np.ndarray]], List[Union[np.ndarray, torch.Tensor]]]:
+        """
+        Scale input data.
+
+        Args:
+            y (Union[pd.DataFrame, np.ndarray, torch.Tensor]): data to scale
+            X (pd.DataFrame): dataframe with ``groups`` columns. Only necessary if :py:class:`~GroupNormalizer`
+                is among normalizers
+            return_norm (bool, optional): If to return . Defaults to False.
+            target_scale (List[torch.Tensor]): target scale to use instead of fitted center and scale
+
+        Returns:
+            Union[List[Tuple[Union[np.ndarray, torch.Tensor], np.ndarray]], List[Union[np.ndarray, torch.Tensor]]]:
+                List of scaled data, if ``return_norm=True``, returns also scales as second element
+        """
+        if isinstance(y, pd.DataFrame):
+            y = y.to_numpy().transpose()
+
+        res = []
+        for idx, normalizer in enumerate(self.normalizers):
+            if target_scale is not None:
+                scale = target_scale[idx]
+            else:
+                scale = None
+            if isinstance(normalizer, GroupNormalizer):
+                r = normalizer.transform(y[idx], X=X, return_norm=return_norm, target_scale=scale)
+            else:
+                r = normalizer.transform(y[idx], return_norm=return_norm, target_scale=scale)
+            res.append(r)
+
+        if return_norm:
+            return [r[0] for r in res], [r[1] for r in res]
+        else:
+            return res
+
+    def __call__(self, data: Dict[str, Union[List[torch.Tensor], torch.Tensor]]) -> List[torch.Tensor]:
+        """
+        Inverse transformation but with network output as input.
+
+        Args:
+            data (Dict[str, Union[List[torch.Tensor], torch.Tensor]]): Dictionary with entries
+                * prediction: list of data to de-scale
+                * target_scale: list of center and scale of data
+
+        Returns:
+            List[torch.Tensor]: list of de-scaled data
+        """
+        denormalized = [
+            normalizer(dict(prediction=data["prediction"][idx], target_scale=data["target_scale"][idx]))
+            for idx, normalizer in enumerate(self.normalizers)
+        ]
+        return denormalized
+
+    def get_parameters(self, *args, **kwargs) -> List[torch.Tensor]:
+        """
+        Returns parameters that were used for encoding.
+
+        Returns:
+            List[torch.Tensor]: First element is center of data and second is scale
+        """
+        return [normalizer.get_parameters(*args, **kwargs) for normalizer in self.normalizers]
+
+    def __getattr__(self, name: str):
+        """
+        Return dynamically attributes.
+
+        Return attributes if defined in this class. If not, create dynamically attributes based on
+        attributes of underlying normalizers that are lists. Create functions if necessary.
+        Arguments to functions are distributed to the functions if they are lists and their length
+        matches the number of normalizers. Otherwise, they are directly passed to each callable of the
+        normalizers.
+
+        Args:
+            name (str): name of attribute
+
+        Returns:
+            attributes of this class or list of attributes of underlying class
+        """
+        try:
+            return super().__getattr__(name)
+        except AttributeError as e:
+            attribute_exists = all([hasattr(norm, name) for norm in self.normalizers])
+            if attribute_exists:
+                # check if to return callable or not and return function if yes
+                if callable(getattr(self.normalizers[0], name)):
+                    n = len(self.normalizers)
+
+                    def func(*args, **kwargs):
+                        # if arg/kwarg is list and of length normalizers, then apply each part to a normalizer.
+                        #  otherwise pass it directly to all normalizers
+                        results = []
+                        for idx, norm in enumerate(self.normalizers):
+                            new_args = [
+                                arg[idx]
+                                if isinstance(arg, (list, tuple))
+                                and not isinstance(arg, rnn.PackedSequence)
+                                and len(arg) == n
+                                else arg
+                                for arg in args
+                            ]
+                            new_kwargs = {
+                                key: val[idx]
+                                if isinstance(val, list) and not isinstance(val, rnn.PackedSequence) and len(val) == n
+                                else val
+                                for key, val in kwargs.items()
+                            }
+                            results.append(getattr(norm, name)(*new_args, **new_kwargs))
+                        return results
+
+                    return func
+                else:
+                    # else return list of attributes
+                    return [getattr(norm, name) for norm in self.normalizers]
+            else:  # attribute does not exist for all normalizers
+                raise e
