@@ -3,7 +3,8 @@ Models for timeseries forecasting.
 """
 from pytorch_forecasting.models.base_model import BaseModel
 from pytorch_forecasting.models.baseline import Baseline
+from pytorch_forecasting.models.deepar import DeepAR
 from pytorch_forecasting.models.nbeats import NBeats
 from pytorch_forecasting.models.temporal_fusion_transformer import TemporalFusionTransformer
 
-__all__ = ["NBeats", "TemporalFusionTransformer", "BaseModel", "Baseline"]
+__all__ = ["NBeats", "TemporalFusionTransformer", "DeepAR", "BaseModel", "Baseline"]
